from aiohttp import web
import asyncio

_default_server_port = 18103


class OneFileHttpServer:
    def __init__(self, log, app, srv, handler, fut):
        self._app = app
        self._srv = srv
        self._handler = handler
        self._log = log
        self._fut = fut

    @classmethod
    @asyncio.coroutine
    def make(cls, log, filename):
        fut = asyncio.Future()

        def receive(request):
            log.debug('Got request')
            yield from request.post()
            try:
                with open(filename, 'wb') as f:
                    g = request.POST['file'].file
                    f.write(g.read())
                fut.set_result(filename)
            except:
                log.exception('Could not receive file')
                fut.set_result(None)

            return web.Response(body=b"Accepted")

        app = web.Application()
        log.debug('Adding POST route at /receive')
        app.router.add_route('POST', '/receive', receive)

        loop = asyncio.get_event_loop()
        handler = app.make_handler()

        # FIXME: this should be tied to the Docker interface,
        # when we have a good way of calculating it
        srv = yield from loop.create_server(handler, '0.0.0.0', _default_server_port)

        return cls(log, app, srv, handler, fut)

    def cancel(self):
        self._fut.cancel()

    @asyncio.coroutine
    def wait(self):
        yield from self._fut
        return self._fut.result()

    @asyncio.coroutine
    def close(self):
        self._srv.close()
        yield from self._srv.wait_closed()
        yield from self._app.shutdown()
        yield from self._handler.finish_connections(60.0)
        yield from self._app.cleanup()



import datetime
import tabulate
import colorama as C
import asyncio
import os
import tarfile
from git import Repo
import shutil
import lxml.etree
import tempfile
import stat
import uuid

import glot.transfer

_repo_locations = {
    'goosefoot': 'https://github.com/go-smart/glossia-container-goosefoot-control'
}


class GlotActor:
    def __init__(self, verbose, force, destination, color):
        self._verbose = verbose
        self._force = force
        self._destination = destination
        self._color = color

    def set_log(self, log):
        self._log = log

    def set_make_call(self, mc):
        self._mc = mc

    @asyncio.coroutine
    def cancel(self, guid):
        log = self._log
        mc = self._mc

        success = yield from mc('cancel', guid.upper())

        if success:
            log.info('Cancelled [%s]' % guid)
        else:
            log.error('Could not cancel [%s]' % guid)

    @asyncio.coroutine
    def launch(self, gssa_xml, tmp_subdirectory, tmp_directory, input_files, definition_files):
        gssa = lxml.etree.parse(gssa_xml)
        log = self._log
        mc = self._mc

        # We tar the definition files into one object for transferring and add
        # it to the definition node
        if definition_files:
            definition_tmp = tempfile.NamedTemporaryFile(suffix='.tar.gz', dir=tmp_directory, delete=False)
            definition_tar = tarfile.open(fileobj=definition_tmp, mode='w:gz')
            for definition_file in definition_files:
                definition_tar.add(definition_file, os.path.basename(definition_file))
                log.debug("Added [%s]" % os.path.basename(definition_file))
            definition_tar.close()
            definition_tmp.close()

            # Note that this makes the file global readable - we assume the
            # parent of the tmp directory is used to control permissions
            os.chmod(definition_tmp.name, stat.S_IROTH | stat.S_IRGRP | stat.S_IRUSR)

            log.debug("Made temporary tar at %s" % definition_tmp.name)
            definition_node = gssa.find('.//definition')
            location_remote = os.path.join('/tmp', 'gssa-transferrer', os.path.basename(definition_tmp.name))
            definition_node.set('location', location_remote)

        # Do the same with the input surfaces
        if input_files:
            input_tmp = tempfile.NamedTemporaryFile(suffix='.tar.gz', dir=tmp_directory, delete=False)
            input_tar = tarfile.open(fileobj=input_tmp, mode='w:gz')
            for input_file in input_files:
                input_tar.add(input_file, os.path.basename(input_file))
                log.debug("Added [%s]" % os.path.basename(input_file))
            input_tar.close()
            input_tmp.close()

            # Note that this makes the file global readable - we assume the
            # parent of the tmp directory is used to control permissions
            os.chmod(input_tmp.name, stat.S_IROTH | stat.S_IRGRP | stat.S_IRUSR)

            log.debug("Made temporary tar at %s" % input_tmp.name)
            input_node = lxml.etree.SubElement(gssa.find('.//transferrer'), 'input')
            location_remote = os.path.join('/tmp', 'gssa-transferrer', os.path.basename(input_tmp.name))
            input_node.set('location', location_remote)

        # Generate a simulation ID
        guid = uuid.uuid1()

        # Run the simulation
        guid = str(guid)
        gssa_string = lxml.etree.tostring(gssa, encoding="unicode")
        yield from mc('init', guid)
        log.info("Initiated...")
        yield from mc('update_settings_xml', guid, gssa_string)
        log.info("Sent XML...")
        yield from mc('finalize', guid, tmp_subdirectory)
        log.info("Finalized settings...")
        yield from mc('start', guid)
        log.info("Started.")

        # These may already have been removed
        for tmp in definition_tmp, input_tmp:
            try:
                os.unlink(tmp.name)
            except OSError:
                pass

    @asyncio.coroutine
    def status(self, guid):
        log = self._log
        mc = self._mc

        simulation = yield from mc('retrieve_status', guid.upper())

        if not simulation:
            log.error('Simulation [%s] not found' % guid)
        else:
            table = simulation.items()
            print(tabulate.tabulate(table))

    @asyncio.coroutine
    def results(self, guid, target, include_diagnostic, inspect_diagnostic):
        log = self._log
        mc = self._mc

        filename = None
        include_diagnostic = include_diagnostic or inspect_diagnostic

        if target is None:
            log.warn(
                "No target given, assuming we should provide "
                "a target for a local Glossia"
            )
            srv = yield from glot.transfer.OneFileHttpServer.make(log, '%s-results.tgz' % guid)
        else:
            srv = None

        success = yield from mc('request_results', guid.upper(), target)

        if not success:
            log.error('Simulation not found')

        if srv is not None:
            if success:
                filename = yield from srv.wait()
            yield from srv.close()

        if not target and not self._destination:
            self._destination = '.'

        destination = self._destination

        if include_diagnostic:
            yield from self._diagnostic(guid, target, inspect_diagnostic)

        if not target and inspect_diagnostic:
            os.makedirs(destination, exist_ok=True)
            with tarfile.open(filename) as f:
                f.extractall(path=destination)

    @asyncio.coroutine
    def search(self, limit, server_limit, sort, guid, fancy=False):
        log = self._log
        mc = self._mc
        color = self._color

        definitions = yield from mc('search', guid.upper() if guid else '', server_limit)

        headers = [
            'GUID',
            'Set Up',
            'Last status',
            '%',
            '',
            'Completed'
        ]

        for d in definitions.values():
            if d['status']:
                try:
                    float(d['status']['percentage'])
                except:
                    d['status']['percentage'] = False

        g, d = "", ""
        table = []
        for g, d in definitions.items():
            if not d:
                raise Exception("Empty definition!")
            try:
                status = d['status'] if d['status'] else {'percentage': False, 'message': None, 'timestamp': None}
                d['status'] = status

                table.append([
                    g,
                    'Y' if d['finalized'] else 'N',
                    '' if not status['timestamp'] else datetime.datetime.fromtimestamp(status['timestamp']).strftime('%A %d, %B %Y :: %H:%M:%S'),
                    '' if not status['percentage'] else ("%.2lf" % status['percentage']),
                    '' if not status['message'] else status['message'].replace('\n', ' ')[0:60],
                    '-' if not d['exit_status'] else ('Y' if d['exit_status'][0] in (True, 'SUCCESS') else 'N')
                ])
            except Exception as e:
                log.error('Could not format status for a simulation')
                log.error(str(e))
                log.error(g)
                # Work around txaio's {} parsing
                log.error(str(d).replace('{', '[').replace('}', ']'))

        if sort == 'timestamp':
            table.sort(key=lambda r: -definitions[r[0]]['status']['timestamp'] if definitions[r[0]]['status']['timestamp'] else 0)
        elif sort == 'guid':
            table.sort(key=lambda r: r[0])

        if limit:
            table = table[:limit]

        if color:
            ce = {'-': C.Fore.YELLOW, 'Y': C.Fore.GREEN, 'N': C.Fore.RED}
            table = [[ce[d[-1]] + di + C.Style.RESET_ALL for di in d] for d in table]

        print(tabulate.tabulate(table, headers=headers, tablefmt=('fancy_grid' if fancy else 'simple')))

    @asyncio.coroutine
    def diagnostic(self, guid, target, inspect):
        log = self._log
        mc = self._mc

        filename = None

        if target is None:
            log.warn(
                "No target given, assuming we should provide "
                "a target for a local Glossia"
            )
            srv = yield from glot.transfer.OneFileHttpServer.make(log, '%s-diagnostic.tgz' % guid)
        else:
            srv = None

        files = yield from mc('request_diagnostic', guid.upper(), target)

        if srv is not None:
            if files:
                filename = yield from srv.wait()
            else:
                srv.cancel()
            srv.close()

        if files:
            log.info("FILES:\n\t%s" % "\n\t".join(["[%s]: [%s]" % t for t in files.items()]))
        else:
            log.warn("No simulation diagnostics found")

        to = self._destination if not target else None
        if inspect:
            log.debug("Inspect")
            if filename is None:
                if len(files) < 1:
                    raise RuntimeError("No diagnostic files returned")
                elif len(files) > 1:
                    raise RuntimeError("Multiple diagnostic files, run inspect manually")

                filename = files.popitem()
                log.debug("Using %s: %s" % filename)
                filename = filename[0]

            with tarfile.open(filename, 'r') as t:
                members = t.getmembers()
                names = [m.name for m in members if not m.isdir()]
                prefix = os.path.commonprefix(names)

            log.debug("Inspecting %s" % filename)
            self._inspect(filename, destination=os.path.join(to, prefix) if to else None)

        return filename

    def inspect(self, archive, destination=None, mode='goosefoot'):
        log = self._log
        verbose = self._verbose
        force = self._force

        if not destination:
            destination = self._destination

        if not os.path.exists(archive):
            raise RuntimeError("You must supply a diagnostic archive")

        if destination:
            path = destination
        else:
            path = os.path.splitext(archive)[0]

        if not force:
            if os.path.exists(path):
                raise FileExistsError("Run with --force to remove existing diagnostic directory (%s)" % path)
        else:
            try:
                shutil.rmtree(path)
            except FileNotFoundError:
                pass

        rootpath = path
        log.debug("Extracting to {path}".format(path=path))

        repo_location = _repo_locations[mode]
        log.debug("Cloning control from {loc}".format(loc=repo_location))
        Repo.clone_from(repo_location, path)

        log.debug("Opening diagnostic archive {arc}".format(arc=archive))
        with tarfile.open(archive, 'r') as t:
            members = t.getmembers()
            names = [m.name for m in members if not m.isdir()]
            prefix = os.path.commonprefix(names)
            log.debug("Stripping prefix {prefix}".format(prefix=prefix))

            inp = os.path.join(prefix, 'input')
            inpfinal = os.path.join(prefix, 'input.final')
            try:
                t.getmember(inp)
            except KeyError:
                try:
                    t.getmember(inpfinal)
                except KeyError:
                    path = os.path.join(path, 'input')
                    os.makedirs(path)

            for m in members:
                if m.name.startswith(prefix):
                    outpath = os.path.join(path, m.name[len(prefix):])
                    outpath = outpath.replace('input.final', 'input')

                    if verbose:
                        log.info("{fm} --> {to}".format(fm=m.name, to=outpath))

                    if m.isdir():
                        os.makedirs(outpath)
                    else:
                        with open(outpath, 'wb') as f, t.extractfile(m) as g:
                            shutil.copyfileobj(g, f)

        log.info("Done extracting")

        if mode is 'goosefoot':
            shutil.copyfile(
                os.path.join(rootpath, 'input', 'settings.xml'),
                os.path.join(rootpath, 'settings', 'settings.xml')
            )



from setuptools import setup

setup(
    name='glot',
    version='0.1.3',
    packages=['glot'],
    package_dir={'glot': 'src/glot'},

    description='CLI manager for Glossia (https://go-smart.github.io/glossia)',
    author='NUMA Engineering Services Ltd.',
    author_email='phil.weir@numa.ie',
    url='http://gosmart-project.eu/',

    scripts=[
        'scripts/glot',
    ],

    install_requires=[
        'aiohttp',
        'txaio',
        'click >= 6.0',
        'gitpython',
        'tabulate',
        'colorama',
        'autobahn'
    ]
)



# This file is part of the Go-Smart Simulation Architecture (GSSA).
# Go-Smart is an EU-FP7 project, funded by the European Commission.
#
# Copyright (C) 2013-  NUMA Engineering Ltd. (see AUTHORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from autobahn.asyncio.wamp import ApplicationSession
from autobahn.asyncio.wamp import ApplicationRunner
import asyncio
import logging
import traceback
from functools import partial

logger = logging.getLogger(__name__)


def execute(action, actor, server, router, port, debug=False, **kwargs):
    responses = []
    if debug:
        logger.info("DEBUG ON")
        logging.getLogger('autobahn').setLevel(logging.DEBUG)

    runner = ApplicationRunner(url="ws://%s:%d/ws" % (router, port), realm="realm1")
    logger.debug("Starting connection")
    runner.run(partial(GlotConnector, responses=responses, action=action, actor=actor, debug=debug, server=server, **kwargs))
    return responses.pop() if responses else None


# This should be adjusted when this issue resolution hits PIP: https://github.com/tavendo/AutobahnPython/issues/332
# http://stackoverflow.com/questions/28293198/calling-a-remote-procedure-from-a-subscriber-and-resolving-the-asyncio-promise
def wrapped_coroutine(f):
    def wrapper(*args, **kwargs):
        coro = f(*args, **kwargs)
        asyncio.async(coro)
    return wrapper
# endSO


# This is the application object for the shell GSSA client
class GlotConnector(ApplicationSession):

    # Accept arguments from the command line
    def __init__(self, x, responses, action, actor, debug, server=None, **kwargs):
        ApplicationSession.__init__(self, x)
        self._kwargs = kwargs
        self._action = action
        self._server = server
        self._responses = responses

        self._actor = actor
        self._actor.set_make_call(self.execute_call)
        self._actor.set_log(self.log)

        self._apis = {}

        if debug:
            # Seemingly the start_logging call is insufficient
            self.log._set_level('trace')
        if server or debug:
            logger.info("Targeting server [%s]" % (server))

    def make_call(self, suffix):
        # If we have a specific server, address it, otherwise we call whichever
        # one got the full namespace
        if self._server:
            return "com.gosmartsimulation.%s.%s" % (self._server, suffix)
        else:
            return "com.gosmartsimulation.%s" % suffix

    @wrapped_coroutine
    @asyncio.coroutine
    def onJoin(self, details):
        logger.debug("Session ready - executing action")

        try:
            self.result = yield from self._action(self._actor, **self._kwargs)
        except Exception as e:
            logging.exception("Problem executing action")
            traceback.print_exc()
            raise e
        finally:
            self.leave()

        logger.debug("Executed")

    @asyncio.coroutine
    def execute_call(self, suffix, *args, minapi='A0.0'):
        if minapi:
            if suffix in self._apis:
                api = self._apis[suffix]
            else:
                try:
                    api = yield from self.call(self.make_call('api'))
                    api = str(api)
                except:
                    api = 'A0.0'
                self._apis[suffix] = api

            isapi, api = api[0], api[1:]

            if isapi != 'A':
                raise RuntimeError('Remote API unknown - %s%s' % (isapi, api))

            api = float(api)
            minapi = float(minapi[1:])
            if api < minapi:
                raise NotImplementedError('The API version of the remote server is too low for this operation (%.1lf < %.1lf)' % (api, minapi))

        try:
            result = yield from self.call(self.make_call(suffix), *args)
            self._responses.append(result)
        except:
            logger.exception("Could not complete call")

        return result

    def onDisconnect(self):
        asyncio.get_event_loop().stop()
